# -*- coding: utf-8 -*-
"""
app.search
~~~~~~~~~~

This module implements the search interface.

:copyright: (c) 2018 European Spallation Source ERIC
:license: BSD 2-Clause, see LICENSE for more details.

"""
from flask import current_app


def add_to_index(index, body):
    """Add a document to an index"""
    if not current_app.elasticsearch:
        return
    # We remove the id key because it's used as the id of the document.
    # Using the model id as document id allows to easily retrieve
    # the model linked to a document
    id = body.pop("id")
    current_app.elasticsearch.index(
        index=index,
        doc_type="_doc",
        id=id,
        body=body,
        refresh=current_app.config["ELASTICSEARCH_REFRESH"],
    )


def remove_from_index(index, id):
    """Remove a document from the index"""
    if not current_app.elasticsearch:
        return
    current_app.elasticsearch.delete(
        index=index,
        doc_type="_doc",
        id=id,
        refresh=current_app.config["ELASTICSEARCH_REFRESH"],
    )


def query_index(index, query, page=1, per_page=20, sort=None):
    """Run a search query on the index

    Return a BaseQuery (with the ids found) and the total number of hits as a tuple
    """
    if not current_app.elasticsearch:
        return [], 0
    kwargs = {
        "index": index,
        "doc_type": "_doc",
        "q": query,
        "from_": (page - 1) * per_page,
        "size": per_page,
    }
    if sort is not None:
        kwargs["sort"] = sort
    search = current_app.elasticsearch.search(**kwargs)
    ids = [int(hit["_id"]) for hit in search["hits"]["hits"]]
    return ids, search["hits"]["total"]
